/// <reference types="node" />
import { IObj } from './object';
import { IEncrypt } from './encrypt';
import { IFont } from "./painter";
import { IBase } from './base-document';
export declare enum NPDFVersion {
    Pdf11 = 0,
    Pdf12 = 1,
    Pdf13 = 2,
    Pdf14 = 3,
    Pdf15 = 4,
    Pdf16 = 5,
    Pdf17 = 6,
}
export declare enum NPDFWriteMode {
    Default = 1,
    Compact = 2,
}
export declare type Callback = (err: Error, data: Buffer | string) => void;
export declare enum NPDFFontEncoding {
    WinAnsi = 1,
    Standard = 2,
    PdfDoc = 3,
    MacRoman = 4,
    MacExpert = 5,
    Symbol = 6,
    ZapfDingbats = 7,
    Win1250 = 8,
    Iso88592 = 9,
    Identity = 0,
}
export declare enum NPDFPageMode {
    DontCare = 0,
    UseNone = 1,
    UseThumbs = 2,
    UseBookmarks = 3,
    FullScreen = 4,
    UseOC = 5,
    UseAttachments = 6,
}
export declare enum NPDFPageLayout {
    Ignore = 0,
    Default = 1,
    SinglePage = 2,
    OneColumn = 3,
    TwoColumnLeft = 4,
    TwoColumnRight = 5,
    TwoPageLeft = 6,
    TwoPageRight = 7,
}
export interface NPDFCreateFontOpts {
    fontName: string;
    bold?: boolean;
    italic?: boolean;
    encoding?: NPDFFontEncoding;
    embed?: boolean;
    fileName?: string;
}
/**
 * @interface IDocument
 * @desc Document represents a PdfMemDocument, construct from an existing pdf document.
 * Document is the core class for reading and manipulating PDF files and writing them back to disk.
 * Document was designed to allow easy access to the object structure of a PDF file.
 * Document should be used whenever you want to change the object structure of a PDF file.
 */
export interface IDocument extends IBase {
    new (): IDocument;
    password: string;
    encrypt: IEncrypt;
    readonly trailer: IObj;
    readonly catalog: IObj;
    load(file: string | Buffer, opts: {
        forUpdate?: boolean;
        fromBuffer?: boolean;
        password?: string;
    }, cb: Callback): void;
    load(file: string | Buffer, cb: Callback): void;
    /**
     * Deletes one or more pages from the document by removing the pages reference
     * from the pages tree. This does NOT remove the page object as the page object
     * may be used by other objects in the document.
     * @param startIndex - first page to delete (0-based)
     * @param count - number of pages to delete
     */
    splicePages(startIndex: number, count: number): void;
    /**
     * Copies one or more pages from another pdf to this document
     * This function copies the entire document to the target document and then
     * deletes pages that are not of interest. This is a much faster process, but
     * without object garbage collection the document may result in a much larger
     * than necessary document
     * @see {IDocument#gc}
     * @param fromDoc - PdfMemDocument to append
     * @param startIndex - first page to copy (0-based)
     * @param count - number of pages to copy
     */
    insertPages(fromDoc: IDocument, startIndex: number, count: number): number;
    /**
     * Persist the document with any changes applied to either a new nodejs buffer or to disk.
     * @param destination - file path or callback function
     * @param cb - if file path was provided as destination, this must be a callback function
     */
    write(destination: Callback | string, cb?: Callback): void;
    /**
     * Looks for a font containg the provided name or id.
     * This operation may be expensive as it iterates all pdf object searching for the font.
     * If nothing is found null is returned
     * @param name - font name or font id
     */
    getFont(name: string): IFont;
    /**
     * Performs garbage collection on the document. All objects not
     * reachable by the trailer or deleted.
     * @param file - pdf document
     * @param pwd - if document is password protected this parameter is required
     * @param output - file location
     * @param cb - function
     */
    gc(file: string, pwd: string, output: string, cb: Callback): void;
}
export declare const documentGc: (file: string, pwd: string, output: string, cb: Callback) => void;
